use super::db::query::GameSaveQuery;
use super::db::Database;
use super::game::{GameFile, GameSaveLocation};
use super::schema::{game_file, game_save_location};
use std::convert::TryInto;
use std::path::PathBuf;
use uuid::Uuid;

#[derive(Queryable)]
pub struct DbGameFile {
    pub id: i32,
    pub original_path: String,
    pub file_hash: Vec<u8>,
    pub game_save_id: i32,
}

impl From<DbGameFile> for GameFile {
    fn from(db_file: DbGameFile) -> Self {
        let original_path = PathBuf::from(db_file.original_path);
        let hash_bytes: [u8; 8] = db_file
            .file_hash
            .try_into()
            .expect("GameFile hash from DB was not 8 bytes long");
        let hash = u64::from_be_bytes(hash_bytes);

        Self {
            original_path,
            hash,
        }
    }
}

impl From<&DbGameFile> for GameFile {
    fn from(db_file: &DbGameFile) -> Self {
        let original_path = PathBuf::from(&db_file.original_path);
        let buf: &[u8] = db_file.file_hash.as_ref();
        let hash_bytes: [u8; 8] = buf
            .try_into()
            .expect("GameFile hash from DB was not 8 bytes long");
        let hash = u64::from_be_bytes(hash_bytes);

        Self {
            original_path,
            hash,
        }
    }
}

#[derive(Insertable)]
#[table_name = "game_file"]
pub struct NewGameFile<'a> {
    pub original_path: &'a str,
    pub file_hash: &'a [u8],
    pub game_save_id: i32,
}

#[derive(Queryable)]
pub struct DbGameSaveLocation {
    pub id: i32,
    pub friendly_name: Option<String>,
    pub original_path: String,
    pub uuid: Vec<u8>,
}

impl From<DbGameSaveLocation> for GameSaveLocation {
    fn from(db_save_loc: DbGameSaveLocation) -> Self {
        // FIXME: This makes .into() and ::from() rather resource intensive
        // This isn't that intuitive. It might be best to abandon the From<> traits here
        // and go for dedicated methods that express how much the database is read from here
        let original_path = PathBuf::from(db_save_loc.original_path);

        let uuid =
            Uuid::from_slice(&db_save_loc.uuid).expect("UUIDv4 from Database was not 16bytes long");
        let conn = super::db::establish_connection();
        let files = Database::get_game_files(&conn, GameSaveQuery::Id(db_save_loc.id))
            .expect("Failed to Interact w/ the Database")
            .expect("Turn this into a TryFrom Please");

        Self {
            friendly_name: db_save_loc.friendly_name,
            original_path,
            files,
            uuid,
        }
    }
}

impl From<&DbGameSaveLocation> for GameSaveLocation {
    fn from(db_save_loc: &DbGameSaveLocation) -> Self {
        // FIXME: See  From<DbGameSaveLocation> for GameSaveLocation
        let original_path = PathBuf::from(&db_save_loc.original_path);

        // FIXME: Is it reasonable to assume that this will **never** fail?
        let uuid =
            Uuid::from_slice(&db_save_loc.uuid).expect("UUIDv4 from Database was not 16bytes long");
        let conn = super::db::establish_connection();
        let files = Database::get_game_files(&conn, GameSaveQuery::Id(db_save_loc.id))
            .expect("Failed to interact w/ the Database")
            .expect("Turn this into a TryFrom please");

        Self {
            friendly_name: db_save_loc.friendly_name.clone(),
            original_path,
            files,
            uuid,
        }
    }
}

#[derive(Insertable)]
#[table_name = "game_save_location"]
pub struct NewGameSaveLocation<'a> {
    pub friendly_name: Option<&'a str>,
    pub original_path: &'a str,
    pub uuid: &'a [u8],
}
